package solatium.blockchain;


import org.springframework.context.annotation.Primary;
import org.springframework.stereotype.Component;

import java.util.*;

@Primary
@Component
public final class BlockChainService implements Chain {

    private final String GENESIS_HASH = "0";
    private final List<HashedBlock> blockChain;

    public BlockChainService() {
        blockChain = new ArrayList<>();
        Transaction genesisTransaction = new Transaction("0", "0", 0);
        Block genesisBlock = new TransactionBlock(GENESIS_HASH, genesisTransaction, 0);
        HashedBlock hashedGenesisBlock = genesisBlock.getHashed();
        blockChain.add(hashedGenesisBlock);
    }

    BlockChainService(List<HashedBlock> blockChain) {
        if (blockChain.isEmpty() || !blockChain.get(0).getBlock().getPreviousHash().equals(GENESIS_HASH)) {
            throw new BlockChainStartsNotWithGenesisBlockException();
        }
        this.blockChain = blockChain;
    }

    @Override
    public void addToChain(HashedBlock block) {
        blockChain.add(block);
    }

    @Override
    public void addToChain(Transaction data) {
        Block block = new TransactionBlock(getLastHash(), data, new Date().getTime());
        addToChain(block.getHashed());
    }

    @Override
    public boolean validateChain() {
        Iterator<HashedBlock> hashedBlockIterator = blockChain.iterator();
        String previousHash = GENESIS_HASH;

        while (hashedBlockIterator.hasNext()) {
            HashedBlock hashedBlock = hashedBlockIterator.next();
            String originalHash = hashedBlock.getHash();
            Block block = hashedBlock.getBlock();
            String newHash = new Sha256Hash(block.getDataToHash(), previousHash).getHash();
            if (!originalHash.equals(newHash) || !block.getPreviousHash().equals(previousHash)) {
                return false;
            }
            previousHash = originalHash;
        }

        return true;
    }

    @Override
    public String getLastHash() {
        return blockChain.get(blockChain.size() - 1).getHash();
    }

    @Override
    public List<HashedBlock> getBlockChain() {
        List<HashedBlock> blockChainList = new ArrayList<>();

        blockChain.iterator().forEachRemaining(blockChainList::add);
        return blockChainList;
    }

    @Override
    public HashedBlock getGenesisBlock() {
        return blockChain.get(0);
    }

    @Override
    public String toString() {
        List<HashedBlock> hashedBlocks = getBlockChain();
        StringBuilder hashedBlocksString = new StringBuilder();
        hashedBlocks.forEach(hashedBlock -> hashedBlocksString.append(hashedBlock.toString()).append("\n"));
        return "Chain{\n" +
                hashedBlocksString.toString() +
                '}';

    }
}
